#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pytorch_lightning as pl
import torch

# TODO: wandb and matplotlib are not in requirements
import matplotlib.pyplot as plt
import wandb

import disent.util.strings.colors as c
from disent.dataset import DisentDataset
from disent.dataset.data import GroundTruthData
from disent.frameworks.ae import Ae
from disent.frameworks.vae import Vae
from disent.util.function import wrapped_partial
from disent.util.iters import chunked
from disent.util.lightning.callbacks._callbacks_base import BaseCallbackPeriodic
from disent.util.lightning.callbacks._helper import _get_dataset_and_ae_like
from disent.util.lightning.logger_util import wb_log_metrics
from disent.util.profiling import Timer
from disent.util.seeds import TempNumpySeed
from disent.util.visualize.plot import plt_subplots_imshow


log = logging.getLogger(__name__)


# ========================================================================= #
# Helper Functions                                                          #
# ========================================================================= #


# helper
def _to_dmat(
    size: int,
    i_a: np.ndarray,
    i_b: np.ndarray,
    dists: Union[torch.Tensor, np.ndarray],
) -> np.ndarray:
    if isinstance(dists, torch.Tensor):
        dists = dists.detach().cpu().numpy()
    # checks
    assert i_a.ndim == 1
    assert i_a.shape == i_b.shape
    assert i_a.shape == dists.shape
    # compute
    dmat = np.zeros([size, size], dtype='float32')
    dmat[i_a, i_b] = dists
    dmat[i_b, i_a] = dists
    return dmat


_AE_DIST_NAMES = ('x', 'z', 'x_recon')
_VAE_DIST_NAMES = ('x', 'z', 'kl', 'x_recon')


@torch.no_grad()
def _get_dists_ae(ae: Ae, x_a: torch.Tensor, x_b: torch.Tensor):
    # feed forware
    z_a, z_b = ae.encode(x_a), ae.encode(x_b)
    r_a, r_b = ae.decode(z_a), ae.decode(z_b)
    # distances
    return [
        ae.recon_handler.compute_pairwise_loss(x_a, x_b),
        torch.norm(z_a - z_b, p=2, dim=-1),  # l2 dist
        ae.recon_handler.compute_pairwise_loss(r_a, r_b),
    ]


@torch.no_grad()
def _get_dists_vae(vae: Vae, x_a: torch.Tensor, x_b: torch.Tensor):
    from torch.distributions import kl_divergence
    # feed forward
    (z_post_a, z_prior_a), (z_post_b, z_prior_b) = vae.encode_dists(x_a), vae.encode_dists(x_b)
    z_a, z_b = z_post_a.mean, z_post_b.mean
    r_a, r_b = vae.decode(z_a), vae.decode(z_b)
    # dists
    kl_ab = 0.5 * kl_divergence(z_post_a, z_post_b) + 0.5 * kl_divergence(z_post_b, z_post_a)
    # distances
    return [
        vae.recon_handler.compute_pairwise_loss(x_a, x_b),
        torch.norm(z_a - z_b, p=2, dim=-1),  # l2 dist
        vae.recon_handler._pairwise_reduce(kl_ab),
        vae.recon_handler.compute_pairwise_loss(r_a, r_b),
    ]


def _get_dists_fn(model: Ae) -> Tuple[Optional[Tuple[str, ...]], Optional[Callable[[object, object], Sequence[Sequence[float]]]]]:
    # get aggregate function
    if isinstance(model, Vae):
        dists_names, dists_fn = _VAE_DIST_NAMES, wrapped_partial(_get_dists_vae, model)
    elif isinstance(model, Ae):
        dists_names, dists_fn = _AE_DIST_NAMES, wrapped_partial(_get_dists_ae, model)
    else:
        dists_names, dists_fn = None, None
    return dists_names, dists_fn


@torch.no_grad()
def _collect_dists_subbatches(dists_fn: Callable[[object, object], Sequence[Sequence[float]]], batch: torch.Tensor, i_a: np.ndarray, i_b: np.ndarray, batch_size: int = 64):
    # feed forward
    results = []
    for idxs in chunked(np.stack([i_a, i_b], axis=-1), chunk_size=batch_size):
        ia, ib = idxs.T
        x_a, x_b = batch[ia], batch[ib]
        # feed forward
        data = dists_fn(x_a, x_b)
        results.append(data)
    return [torch.cat(r, dim=0) for r in zip(*results)]


def _compute_and_collect_dists(
    dataset: DisentDataset,
    dists_fn,
    dists_names: Sequence[str],
    traversal_repeats: int = 100,
    batch_size: int = 32,
    include_gt_factor_dists: bool = True,
    transform_batch: Callable[[object], object] = None,
    data_mode: str = 'input',
) -> Tuple[Tuple[str, ...], List[List[np.ndarray]]]:
    assert traversal_repeats > 0
    gt_data = dataset.gt_data
    # generate
    f_grid = []
    # generate
    for f_idx, f_size in enumerate(gt_data.factor_sizes):
        # save for the current factor (traversal_repeats, len(names), len(i_a))
        f_dists = []
        # upper triangle excluding diagonal
        i_a, i_b = np.triu_indices(f_size, k=1)
        # repeat over random traversals
        for i in range(traversal_repeats):
            # get random factor traversal
            factors = gt_data.sample_random_factor_traversal(f_idx=f_idx)
            indices = gt_data.pos_to_idx(factors)
            # load data
            batch = dataset.dataset_batch_from_indices(indices, data_mode)
            if transform_batch is not None:
                batch = transform_batch(batch)
            # feed forward & compute dists -- (len(names), len(i_a))
            dists = _collect_dists_subbatches(dists_fn=dists_fn, batch=batch, i_a=i_a, i_b=i_b, batch_size=batch_size)
            assert len(dists) == len(dists_names)
            # distances
            f_dists.append(dists)
        # aggregate all dists into distances matrices for current factor
        f_dmats = [
            _to_dmat(size=f_size, i_a=i_a, i_b=i_b, dists=torch.stack(dists, dim=0).mean(dim=0))
            for dists in zip(*f_dists)
        ]
        # handle factors
        if include_gt_factor_dists:
            i_dmat = _to_dmat(size=f_size, i_a=i_a, i_b=i_b, dists=np.abs(factors[i_a] - factors[i_b]).sum(axis=-1))
            f_dmats = [i_dmat, *f_dmats]
        # append data
        f_grid.append(f_dmats)
    # handle factors
    if include_gt_factor_dists:
        dists_names = ('factors', *dists_names)
    # done
    return tuple(dists_names), f_grid


def compute_factor_distances(
    dataset: DisentDataset,
    dists_fn,
    dists_names: Sequence[str],
    traversal_repeats: int = 100,
    batch_size: int = 32,
    include_gt_factor_dists: bool = True,
    transform_batch: Callable[[object], object] = None,
    seed: Optional[int] = 777,
    data_mode: str = 'input',
) -> Tuple[Tuple[str, ...], List[List[np.ndarray]]]:
    # log this callback
    gt_data = dataset.gt_data
    log.info(f'| {gt_data.name} - computing factor distances...')
    # compute various distances matrices for each factor
    with Timer() as timer, TempNumpySeed(seed):
        dists_names, f_grid = _compute_and_collect_dists(
            dataset=dataset,
            dists_fn=dists_fn,
            dists_names=dists_names,
            traversal_repeats=traversal_repeats,
            batch_size=batch_size,
            include_gt_factor_dists=include_gt_factor_dists,
            transform_batch=transform_batch,
            data_mode=data_mode,
        )
    # log this callback!
    log.info(f'| {gt_data.name} - computed factor distances! time{c.GRY}={c.lYLW}{timer.pretty:<9}{c.RST}')
    return dists_names, f_grid


def plt_factor_distances(
    gt_data: GroundTruthData,
    f_grid: List[List[np.ndarray]],
    dists_names: Sequence[str],
    title: str,
    plt_block_size: float = 1.25,
    plt_transpose: bool = False,
    plt_cmap='Blues',
):
    # plot information
    imshow_kwargs = dict(cmap=plt_cmap)
    figsize       = (plt_block_size*len(f_grid[0]), plt_block_size * gt_data.num_factors)
    # plot!
    if not plt_transpose:
        fig, axs = plt_subplots_imshow(grid=f_grid,             col_labels=dists_names,          row_labels=gt_data.factor_names, figsize=figsize,       title=title, imshow_kwargs=imshow_kwargs)
    else:
        fig, axs = plt_subplots_imshow(grid=list(zip(*f_grid)), col_labels=gt_data.factor_names, row_labels=dists_names,          figsize=figsize[::-1], title=title, imshow_kwargs=imshow_kwargs)
    # done
    return fig, axs


# ========================================================================= #
# Data Dists Visualisation Callback                                         #
# ========================================================================= #


class VaeGtDistsLoggingCallback(BaseCallbackPeriodic):

    def __init__(
        self,
        seed: Optional[int] = 7777,
        every_n_steps: Optional[int] = None,
        traversal_repeats: int = 100,
        begin_first_step: bool = False,
        plt_block_size: float = 1.25,
        plt_show: bool = False,
        plt_transpose: bool = False,
        log_wandb: bool = True,  # TODO: detect this automatically?
        batch_size: int = 128,
        include_factor_dists: bool = True,
    ):
        assert traversal_repeats > 0
        self._traversal_repeats = traversal_repeats
        self._seed = seed
        self._plt_block_size = plt_block_size
        self._plt_show = plt_show
        self._log_wandb = log_wandb
        self._include_gt_factor_dists = include_factor_dists
        self._transpose_plot = plt_transpose
        self._batch_size = batch_size
        super().__init__(every_n_steps, begin_first_step)

    @torch.no_grad()
    def do_step(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        # exit early
        if not (self._plt_show or self._log_wandb):
            log.warning(f'skipping {self.__class__.__name__} neither `plt_show` or `log_wandb` is `True`!')
            return
        # get dataset and vae framework from trainer and module
        dataset, vae = _get_dataset_and_ae_like(trainer, pl_module, unwrap_groundtruth=True)
        # exit early
        if not dataset.is_ground_truth:
            log.warning(f'cannot run {self.__class__.__name__} over non-ground-truth data, skipping!')
            return
        # get aggregate function
        dists_names, dists_fn = _get_dists_fn(vae)
        if (dists_names is None) or (dists_fn is None):
            log.warning(f'cannot run {self.__class__.__name__}, unsupported model type: {type(vae)}, must be {Ae.__name__} or {Vae.__name__}')
            return
        # compute various distances matrices for each factor
        dists_names, f_grid = compute_factor_distances(
            dataset=dataset,
            dists_fn=dists_fn,
            dists_names=dists_names,
            traversal_repeats=self._traversal_repeats,
            batch_size=self._batch_size,
            include_gt_factor_dists=self._include_gt_factor_dists,
            transform_batch=lambda batch: batch.to(vae.device),
            seed=self._seed,
            data_mode='input',
        )
        # plot these results
        fig, axs = plt_factor_distances(
            gt_data=dataset.gt_data,
            f_grid=f_grid,
            dists_names=dists_names,
            title=f'{vae.__class__.__name__}: {dataset.gt_data.name.capitalize()} Distances',
            plt_block_size=self._plt_block_size,
            plt_transpose=self._transpose_plot,
            plt_cmap='Blues',
        )
        # show the plot
        if self._plt_show:
            plt.show()
        # log the plot to wandb
        if self._log_wandb:
            wb_log_metrics(trainer.logger, {
                'factor_distances': wandb.Image(fig)
            })


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import final
from typing import Optional
from typing import Sequence

import torch
from torch.distributions import Normal

from disent.frameworks.helper.reconstructions import make_reconstruction_loss
from disent.frameworks.helper.reconstructions import ReconLossHandler
from disent.frameworks.vae._supervised__adaneg_tvae import AdaNegTripletVae
from disent.nn.loss.triplet_mining import configured_idx_mine


log = logging.getLogger(__name__)


# ========================================================================= #
# Mixin                                                                     #
# ========================================================================= #


class DataOverlapMixin(object):

    # should be inherited by the config on the child class
    @dataclass
    class cfg:
        # override from AE
        recon_loss: str = 'mse'
        # OVERLAP VAE
        overlap_loss: Optional[str] = None  # if None, use the value from recon_loss
        overlap_num: int = 1024
        overlap_mine_ratio: float = 0.1
        overlap_mine_triplet_mode: str = 'none'
        # AUGMENT
        overlap_augment_mode: str = 'augment'
        overlap_augment: Optional[dict] = None

    # private properties
    # - since this class does not have a constructor, it
    #   provides the `init_data_overlap_mixin` method, which
    #   should be called inside the constructor of the child class
    _augment: callable
    _overlap_handler: ReconLossHandler
    _init: bool

    def init_data_overlap_mixin(self):
        if hasattr(self, '_init'):
            raise RuntimeError(f'{DataOverlapMixin.__name__} on {self.__class__.__name__} was initialised more than once!')
        self._init = True
        # set augment and instantiate if needed
        if self.cfg.overlap_augment is not None:
            import hydra
            self._augment = hydra.utils.instantiate(self.cfg.overlap_augment)
            assert (self._augment is None) or callable(self._augment), f'augment is not None or callable: {repr(self._augment)}, obtained from `overlap_augment={repr(self.cfg.overlap_augment)}`'
        else:
            self._augment = None
        # get overlap loss
        if self.cfg.overlap_loss is None:
            log.info(f'`overlap_loss` not specified for {repr(self.__class__.__name__)}, using `recon_loss` instead: {repr(self.cfg.recon_loss)}')
            overlap_loss = self.cfg.recon_loss
        else:
            overlap_loss = self.cfg.overlap_loss
        # construct the overlap handler
        self._overlap_handler: ReconLossHandler = make_reconstruction_loss(name=overlap_loss, reduction='mean')

    @final
    @property
    def overlap_handler(self) -> ReconLossHandler:
        return self._overlap_handler

    def overlap_swap_triplet_idxs(self, x_targ, a_idxs, p_idxs, n_idxs):
        xs_targ = [x_targ[idxs] for idxs in (a_idxs, p_idxs, n_idxs)]
        # CORE: order the latent variables for triplet
        swap_mask = self.overlap_swap_mask(xs_targ=xs_targ)
        # swap all idxs
        swapped_a_idxs = a_idxs
        swapped_p_idxs = torch.where(swap_mask, n_idxs, p_idxs)
        swapped_n_idxs = torch.where(swap_mask, p_idxs, n_idxs)
        # return values
        return swapped_a_idxs, swapped_p_idxs, swapped_n_idxs

    def overlap_swap_mask(self, xs_targ: Sequence[torch.Tensor]) -> torch.Tensor:
        # get variables
        a_x_targ_OLD, p_x_targ_OLD, n_x_targ_OLD = xs_targ
        # CORE OF THIS APPROACH
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # calculate which are wrong!
        # TODO: add more loss functions, like perceptual & others
        with torch.no_grad():
            a_p_losses = self.overlap_handler.compute_pairwise_loss(a_x_targ_OLD, p_x_targ_OLD)  # (B, C, H, W) -> (B,)
            a_n_losses = self.overlap_handler.compute_pairwise_loss(a_x_targ_OLD, n_x_targ_OLD)  # (B, C, H, W) -> (B,)
            swap_mask = (a_p_losses > a_n_losses)  # (B,)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        return swap_mask

    @torch.no_grad()
    def augment_batch(self, x_targ):
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # perform the augments
        if self.cfg.overlap_augment_mode in ('augment', 'augment_each'):
            # recreate augment each time
            if self.cfg.overlap_augment_mode == 'augment_each':
                import hydra
                self._augment = hydra.utils.instantiate(self.cfg.overlap_augment)
            # augment on correct device, but skip if not defined!
            aug_x_targ = x_targ if (self._augment is None) else self._augment(x_targ)
        elif self.cfg.overlap_augment_mode == 'none':
            # no augment
            aug_x_targ = x_targ
        else:
            raise KeyError(f'invalid cfg.overlap_augment_mode={repr(self.cfg.overlap_augment_mode)}')
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # checks
        assert x_targ.shape == aug_x_targ.shape
        return aug_x_targ

    def mine_triplets(self, x_targ, a_idxs, p_idxs, n_idxs):
        return configured_idx_mine(
            x_targ=x_targ,
            a_idxs=a_idxs,
            p_idxs=p_idxs,
            n_idxs=n_idxs,
            cfg=self.cfg,
            pairwise_loss_fn=self.overlap_handler.compute_pairwise_loss,
        )

    def random_mined_triplets(self, x_targ_orig: torch.Tensor):
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # 1. augment batch
        aug_x_targ = self.augment_batch(x_targ_orig)
        # 2. generate random triples -- this does not generate unique pairs
        a_idxs, p_idxs, n_idxs = torch.randint(len(aug_x_targ), size=(3, min(self.cfg.overlap_num, len(aug_x_targ)**3)), device=aug_x_targ.device)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # self.debug(x_targ_orig, x_targ, a_idxs, p_idxs, n_idxs)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # TODO: this can be merged into a single function -- inefficient currently with deltas computed twice
        # 3. reorder random triples
        a_idxs, p_idxs, n_idxs = self.overlap_swap_triplet_idxs(aug_x_targ, a_idxs, p_idxs, n_idxs)
        # 4. mine random triples
        a_idxs, p_idxs, n_idxs = self.mine_triplets(aug_x_targ, a_idxs, p_idxs, n_idxs)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        return a_idxs, p_idxs, n_idxs

    # def debug(self, x_targ_orig, x_targ, a_idxs, p_idxs, n_idxs):
    #     a_p_overlap_orig = - self.recon_handler.compute_unreduced_loss(x_targ_orig[a_idxs], x_targ_orig[p_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_n_overlap_orig = - self.recon_handler.compute_unreduced_loss(x_targ_orig[a_idxs], x_targ_orig[n_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_p_overlap = - self.recon_handler.compute_unreduced_loss(x_targ[a_idxs], x_targ[p_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_n_overlap = - self.recon_handler.compute_unreduced_loss(x_targ[a_idxs], x_targ[n_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_p_overlap_mul = - (a_p_overlap_orig * a_p_overlap)
    #     a_n_overlap_mul = - (a_n_overlap_orig * a_n_overlap)
    #     # check number of things
    #     (up_values_orig, up_counts_orig) = torch.unique(a_p_overlap_orig, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values_orig, un_counts_orig) = torch.unique(a_n_overlap_orig, sorted=True, return_inverse=False, return_counts=True)
    #     (up_values, up_counts) = torch.unique(a_p_overlap, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values, un_counts) = torch.unique(a_n_overlap, sorted=True, return_inverse=False, return_counts=True)
    #     (up_values_mul, up_counts_mul) = torch.unique(a_p_overlap_mul, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values_mul, un_counts_mul) = torch.unique(a_n_overlap_mul, sorted=True, return_inverse=False, return_counts=True)
    #     # plot!
    #     plt.scatter(up_values_orig.detach().cpu(), torch.cumsum(up_counts_orig, dim=-1).detach().cpu())
    #     plt.scatter(un_values_orig.detach().cpu(), torch.cumsum(un_counts_orig, dim=-1).detach().cpu())
    #     plt.scatter(up_values.detach().cpu(), torch.cumsum(up_counts, dim=-1).detach().cpu())
    #     plt.scatter(un_values.detach().cpu(), torch.cumsum(un_counts, dim=-1).detach().cpu())
    #     plt.scatter(up_values_mul.detach().cpu(), torch.cumsum(up_counts_mul, dim=-1).detach().cpu())
    #     plt.scatter(un_values_mul.detach().cpu(), torch.cumsum(un_counts_mul, dim=-1).detach().cpu())
    #     plt.show()
    #     time.sleep(10)


# ========================================================================= #
# Data Overlap Triplet VAE                                                  #
# ========================================================================= #


class DataOverlapTripletVae(AdaNegTripletVae, DataOverlapMixin):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is the unsupervised version of the Adaptive Triplet Loss for Variational Auto-Encoders
    - Triplets are sampled from training minibatches and are ordered using some distance
      function taken directly over the datapoints.
    - Adaptive Triplet Loss is used to guide distance learning and encourage disentanglement
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(AdaNegTripletVae.cfg, DataOverlapMixin.cfg):
        pass

    def __init__(self, model: 'AutoEncoder', cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # initialise mixin
        self.init_data_overlap_mixin()

    def hook_compute_ave_aug_loss(self, ds_posterior: Sequence[Normal], ds_prior, zs_sampled, xs_partial_recon, xs_targ: Sequence[torch.Tensor]):
        [d_posterior], [x_targ_orig] = ds_posterior, xs_targ
        # 1. randomly generate and mine triplets using augmented versions of the inputs
        a_idxs, p_idxs, n_idxs = self.random_mined_triplets(x_targ_orig=x_targ_orig)
        # 2. compute triplet loss
        loss, loss_log = AdaNegTripletVae.estimate_ada_triplet_loss(
            ds_posterior=[Normal(d_posterior.loc[idxs], d_posterior.scale[idxs]) for idxs in (a_idxs, p_idxs, n_idxs)],
            cfg=self.cfg,
        )
        return loss, {
            **loss_log,
        }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import Dict
from typing import final
from typing import Tuple

import torch

from disent.frameworks import DisentFramework
from disent.frameworks.helper.reconstructions import make_reconstruction_loss
from disent.frameworks.helper.reconstructions import ReconLossHandler
from disent.model import AutoEncoder


log = logging.getLogger(__name__)


# ========================================================================= #
# framework_vae                                                             #
# ========================================================================= #


class _AeAndVaeMixin(DisentFramework):
    """
    Base class containing common logic for both the Ae and Vae classes.
    -- This is private because handling of both classes needs to be conducted differently.
       An instance of a Vae should not be an instance of an Ae and vice-versa
    """

    @dataclass
    class cfg(DisentFramework.cfg):
        recon_loss: str = 'mse'
        # multiple reduction modes exist for the various loss components.
        # - 'sum': sum over the entire batch
        # - 'mean': mean over the entire batch
        # - 'mean_sum': sum each observation, returning the mean sum over the batch
        loss_reduction: str = 'mean'
        # disable various components
        detach_decoder: bool = False
        disable_rec_loss: bool = False
        disable_aug_loss: bool = False

    # --------------------------------------------------------------------- #
    # AE/VAE Attributes                                                     #
    # --------------------------------------------------------------------- #

    @property
    def REQUIRED_Z_MULTIPLIER(self) -> int:
        raise NotImplementedError

    @property
    def REQUIRED_OBS(self) -> int:
        raise NotImplementedError

    @final
    @property
    def recon_handler(self) -> ReconLossHandler:
        return self.__recon_handler

    # --------------------------------------------------------------------- #
    # AE/VAE Init                                                           #
    # --------------------------------------------------------------------- #

    # attributes provided by this class and initialised in _init_ae_mixin
    _model: AutoEncoder
    __recon_handler: ReconLossHandler

    def _init_ae_mixin(self, model: AutoEncoder):
        # vae model
        self._model = model
        # check the model
        assert isinstance(self._model, AutoEncoder), f'model must be an instance of {AutoEncoder.__name__}, got: {type(model)}'
        assert self._model.z_multiplier == self.REQUIRED_Z_MULTIPLIER, f'model z_multiplier is {repr(self._model.z_multiplier)} but {self.__class__.__name__} requires that it is: {repr(self.REQUIRED_Z_MULTIPLIER)}'
        # recon loss & activation fn
        self.__recon_handler: ReconLossHandler = make_reconstruction_loss(self.cfg.recon_loss, reduction=self.cfg.loss_reduction)

    # --------------------------------------------------------------------- #
    # AE/VAE Training Step Helper                                           #
    # --------------------------------------------------------------------- #

    @final
    def _get_xs_and_targs(self, batch: Dict[str, Tuple[torch.Tensor, ...]], batch_idx) -> Tuple[Tuple[torch.Tensor, ...], Tuple[torch.Tensor, ...]]:
        xs_targ = batch['x_targ']
        if 'x' not in batch:
            # TODO: re-enable this warning but only ever print once!
            # warnings.warn('dataset does not have input: x -> x_targ using target as input: x_targ -> x_targ')
            xs = xs_targ
        else:
            xs = batch['x']
        # check that we have the correct number of inputs
        if (len(xs) != self.REQUIRED_OBS) or (len(xs_targ) != self.REQUIRED_OBS):
            log.warning(f'batch len(xs)={len(xs)} and len(xs_targ)={len(xs_targ)} observation count mismatch, requires: {self.REQUIRED_OBS}')
        # done
        return xs, xs_targ

    # --------------------------------------------------------------------- #
    # AE/VAE Model Utility Functions (Visualisation)                        #
    # --------------------------------------------------------------------- #

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """Get the deterministic latent representation (useful for visualisation)"""
        raise NotImplementedError

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into reconstruction x_recon (useful for visualisation)"""
        raise NotImplementedError

    def forward(self, batch: torch.Tensor) -> torch.Tensor:
        """Feed through the full deterministic model (useful for visualisation)"""
        raise NotImplementedError

    # --------------------------------------------------------------------- #
    # AE/VAE Model Utility Functions (Training)                             #
    # --------------------------------------------------------------------- #

    def decode_partial(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into partial reconstructions that exclude the final activation if there is one."""
        raise NotImplementedError


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import torch
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution
from torch.distributions import Distribution


# ========================================================================= #
# Kl Losses                                                                 #
# ========================================================================= #


def kl_loss_direct_reverse(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # This is how the original VAE/BetaVAE papers do it.
    # - we compute the reverse kl divergence directly instead of approximating it
    # - kl(post|prior)
    # FORWARD vs. REVERSE kl (https://www.tuananhle.co.uk/notes/reverse-forward-kl.html)
    # - If we minimize the kl(post|prior) or the reverse/exclusive KL, the zero-forcing/mode-seeking behavior arises.
    # - If we minimize the kl(prior|post) or the forward/inclusive KL, the mass-covering/mean-seeking behavior arises.
    return torch.distributions.kl_divergence(posterior, prior)


def kl_loss_approx_reverse(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # This is how pytorch-lightning-bolts does it:
    # - kl(post|prior)
    # See issue: https://github.com/PyTorchLightning/pytorch-lightning-bolts/issues/565
    # - we approximate the reverse kl divergence instead of computing it analytically
    assert z_sampled is not None, 'to compute the approximate kl loss, z_sampled needs to be defined (cfg.kl_mode="approx")'
    return posterior.log_prob(z_sampled) - prior.log_prob(z_sampled)


def kl_loss_direct_forward(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the forward kl
    # - kl(prior|post)
    return torch.distributions.kl_divergence(prior, posterior)


def kl_loss_approx_forward(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the approximate forward kl
    # - kl(prior|post)
    assert z_sampled is not None, 'to compute the approximate kl loss, z_sampled needs to be defined (cfg.kl_mode="approx")'
    return prior.log_prob(z_sampled) - posterior.log_prob(z_sampled)


def kl_loss_direct_symmetric(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the (scaled) symmetric kl
    # - 0.5 * kl(prior|post) + 0.5 * kl(prior|post)
    return 0.5 * kl_loss_direct_reverse(posterior, prior, z_sampled) + 0.5 * kl_loss_direct_forward(posterior, prior, z_sampled)


def kl_loss_approx_symmetric(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the approximate (scaled) symmetric kl
    # - 0.5 * kl(prior|post) + 0.5 * kl(prior|post)
    return 0.5 * kl_loss_approx_reverse(posterior, prior, z_sampled) + 0.5 * kl_loss_approx_forward(posterior, prior, z_sampled)


_KL_LOSS_MODES = {
    # reverse kl -- how it should be done for VAEs
    'direct':         kl_loss_direct_reverse,  # alias for reverse modes
    'approx':         kl_loss_approx_reverse,  # alias for reverse modes
    'direct_reverse': kl_loss_direct_reverse,
    'approx_reverse': kl_loss_approx_reverse,
    # forward kl
    'direct_forward': kl_loss_direct_forward,
    'approx_forward': kl_loss_approx_forward,
    # symmetric kl
    'direct_symmetric': kl_loss_direct_symmetric,
    'approx_symmetric': kl_loss_approx_symmetric,
}


def kl_loss(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None, mode='direct'):
    return _KL_LOSS_MODES[mode](posterior, prior, z_sampled)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple

import torch
from dataclasses import dataclass
from torch.distributions import Distribution
from torch.distributions import kl_divergence
from torch.distributions import Normal

from disent.frameworks.vae._unsupervised__betavae import BetaVae


# ========================================================================= #
# Ada-GVAE                                                                  #
# ========================================================================= #


class AdaVae(BetaVae):
    """
    Weakly Supervised Disentanglement Learning Without Compromises: https://arxiv.org/abs/2002.02886
    - pretty much a beta-vae with averaging between decoder outputs to form weak supervision signal.
    - GAdaVAE:   Averaging from https://arxiv.org/abs/1809.02383
    - ML-AdaVAE: Averaging from https://arxiv.org/abs/1705.08841

    MODIFICATION:
    - Symmetric KL Calculation used by default, described in: https://arxiv.org/pdf/2010.14407.pdf
    - adjustable threshold value

    * This class is a little over complicated because it has the added functionality
      listed above, and because we want to re-use features elsewhere. The code can
      be compressed down into about ~20 neat lines for `hook_intercept_ds` if we
      select and chose fixed `cfg` values.
    """

    REQUIRED_OBS = 2

    @dataclass
    class cfg(BetaVae.cfg):
        ada_average_mode: str = 'gvae'
        ada_thresh_mode: str = 'symmetric_kl'  # kl, symmetric_kl, dist, sampled_dist
        ada_thresh_ratio: float = 0.5

    def hook_intercept_ds(self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution]) -> Tuple[Sequence[Distribution], Sequence[Distribution], Dict[str, Any]]:
        """
        Adaptive VAE Glue Method, putting the various components together
        1. find differences between deltas
        2. estimate a threshold for differences
        3. compute a shared mask from this threshold
        4. average together elements that should be considered shared

        (✓) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
        """
        d0_posterior, d1_posterior = ds_posterior
        # shared elements that need to be averaged, computed per pair in the batch.
        share_mask = self.compute_shared_mask_from_posteriors(d0_posterior, d1_posterior, thresh_mode=self.cfg.ada_thresh_mode, ratio=self.cfg.ada_thresh_ratio)
        # compute average posteriors
        new_ds_posterior = self.make_shared_posteriors(d0_posterior, d1_posterior, share_mask, average_mode=self.cfg.ada_average_mode)
        # return new args & generate logs
        return new_ds_posterior, ds_prior, {
            'shared': share_mask.sum(dim=1).float().mean()
        }

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - POSTERIORS                                                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_deltas_from_posteriors(cls, d0_posterior: Distribution, d1_posterior: Distribution, thresh_mode: str):
        """
        (✓) Visual inspection against reference implementation
        https://github.com/google-research/disentanglement_lib (compute_kl)
        - difference is that they don't multiply by 0.5 to get true kl, but that's not needed

        TODO: this might be numerically unstable with f32 passed to distributions
        """
        # shared elements that need to be averaged, computed per pair in the batch.
        # [𝛿_i ...]
        if thresh_mode == 'kl':
            # ORIGINAL
            deltas = kl_divergence(d1_posterior, d0_posterior)
        elif thresh_mode == 'symmetric_kl':
            # FROM: https://openreview.net/pdf?id=8VXvj1QNRl1
            kl_deltas_d1_d0 = kl_divergence(d1_posterior, d0_posterior)
            kl_deltas_d0_d1 = kl_divergence(d0_posterior, d1_posterior)
            deltas = (0.5 * kl_deltas_d1_d0) + (0.5 * kl_deltas_d0_d1)
        elif thresh_mode == 'dist':
            deltas = cls.compute_deltas_from_zs(d1_posterior.mean, d0_posterior.mean)
        elif thresh_mode == 'sampled_dist':
            deltas = cls.compute_deltas_from_zs(d1_posterior.rsample(), d0_posterior.rsample())
        else:
            raise KeyError(f'invalid thresh_mode: {repr(thresh_mode)}')

        # return values
        return deltas

    @classmethod
    def compute_shared_mask_from_posteriors(cls, d0_posterior: Distribution, d1_posterior: Distribution, thresh_mode: str, ratio=0.5):
        return cls.estimate_shared_mask(z_deltas=cls.compute_deltas_from_posteriors(d0_posterior, d1_posterior, thresh_mode=thresh_mode), ratio=ratio)

    @classmethod
    def make_shared_posteriors(cls, d0_posterior: Normal, d1_posterior: Normal, share_mask: torch.Tensor, average_mode: str) -> Tuple[Normal, Normal]:
        # compute average posterior
        ave_posterior = AdaVae.compute_average_distribution(d0_posterior=d0_posterior, d1_posterior=d1_posterior, average_mode=average_mode)
        # select shared elements
        ave_d0_posterior = Normal(loc=torch.where(share_mask, ave_posterior.loc, d0_posterior.loc), scale=torch.where(share_mask, ave_posterior.scale, d0_posterior.scale))
        ave_d1_posterior = Normal(loc=torch.where(share_mask, ave_posterior.loc, d1_posterior.loc), scale=torch.where(share_mask, ave_posterior.scale, d1_posterior.scale))
        # return values
        return ave_d0_posterior, ave_d1_posterior

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - MEAN/MU VALUES (same functionality as posterior versions)    #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_deltas_from_zs(cls, z0: torch.Tensor, z1: torch.Tensor) -> torch.Tensor:
        return torch.abs(z0 - z1)

    @classmethod
    def compute_shared_mask_from_zs(cls, z0: torch.Tensor, z1: torch.Tensor, ratio: float = 0.5):
        return cls.estimate_shared_mask(z_deltas=cls.compute_deltas_from_zs(z0, z1), ratio=ratio)

    @classmethod
    def make_shared_zs(cls, z0: torch.Tensor, z1: torch.Tensor, share_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        # compute average values
        ave = 0.5 * z0 + 0.5 * z1
        # select shared elements
        ave_z0 = torch.where(share_mask, ave, z0)
        ave_z1 = torch.where(share_mask, ave, z1)
        return ave_z0, ave_z1

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - COMMON                                                       #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def estimate_shared_mask(cls, z_deltas: torch.Tensor, ratio: float) -> torch.Tensor:
        """
        Core of the adaptive VAE algorithm, estimating which factors
        have changed (or in this case which are shared and should remained unchanged
        by being be averaged) between pairs of observations.
        - custom ratio is an addition, when ratio==0.5 then
          this is equivalent to the original implementation.

        (✓) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
            - Implementation conversion is non-trivial, items are histogram binned.
              If we are in the second histogram bin, ie. 1, then kl_deltas <= kl_threshs
            - TODO: (aggregate_labels) An alternative mode exists where you can bind the
                    latent variables to any individual label, by one-hot encoding which
                    latent variable should not be shared: "enforce that each dimension
                    of the latent code learns one factor (dimension 1 learns factor 1)
                    and enforce that each factor of variation is encoded in a single
                    dimension."
        """
        assert 0 <= ratio <= 1, f'ratio must be in the range: 0 <= ratio <= 1, got: {repr(ratio)}'
        # threshold τ
        maximums = z_deltas.max(axis=1, keepdim=True).values      # (B, 1)
        minimums = z_deltas.min(axis=1, keepdim=True).values      # (B, 1)
        z_threshs = torch.lerp(minimums, maximums, weight=ratio)  # (B, 1)
        # true if 'unchanged' and should be average
        shared_mask = z_deltas < z_threshs                        # broadcast (B, Z) and (B, 1) -> (B, Z)
        # return
        return shared_mask

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # HELPER - DISTRIBUTIONS                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def compute_average_distribution(cls, d0_posterior: Normal, d1_posterior: Normal, average_mode: str) -> Normal:
        return _COMPUTE_AVE_FNS[average_mode](
            d0_posterior=d0_posterior,
            d1_posterior=d1_posterior,
        )


# ========================================================================= #
# Averaging Functions                                                       #
# ========================================================================= #


def compute_average_gvae_std(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the arithmetic mean of the encoder distributions.
    - This is a custom function based on the Ada-GVAE averaging,
      except over the standard deviation instead of the variance!

    *NB* this is un-official!
    """
    assert isinstance(d0_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}'
    assert isinstance(d1_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}'
    # averages
    ave_std = 0.5 * (d0_posterior.stddev + d1_posterior.stddev)
    ave_mean = 0.5 * (d0_posterior.mean + d1_posterior.mean)
    # done!
    return Normal(loc=ave_mean, scale=ave_std)


def compute_average_gvae(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the arithmetic mean of the encoder distributions.
    - Ada-GVAE Averaging function

    (✓) Visual inspection against reference implementation:
        https://github.com/google-research/disentanglement_lib (GroupVAEBase.model_fn)
    """
    assert isinstance(d0_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}'
    assert isinstance(d1_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}'
    # averages
    ave_var = 0.5 * (d0_posterior.variance + d1_posterior.variance)
    ave_mean = 0.5 * (d0_posterior.mean + d1_posterior.mean)
    # done!
    return Normal(loc=ave_mean, scale=torch.sqrt(ave_var))


def compute_average_ml_vae(d0_posterior: Normal, d1_posterior: Normal) -> Normal:
    """
    Compute the product of the encoder distributions.
    - Ada-ML-VAE Averaging function

    (✓) Visual inspection against reference implementation:
        https://github.com/google-research/disentanglement_lib (MLVae.model_fn)

    # TODO: recheck
    """
    assert isinstance(d0_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}'
    assert isinstance(d1_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}'
    # Diagonal matrix inverse: E^-1 = 1 / E
    # https://proofwiki.org/wiki/Inverse_of_Diagonal_Matrix
    z0_invvar, z1_invvar = d0_posterior.variance.reciprocal(), d1_posterior.variance.reciprocal()
    # average var: E^-1 = E1^-1 + E2^-1
    # disentanglement_lib: ave_var = 2 * z0_var * z1_var / (z0_var + z1_var)
    ave_var = 2 * (z0_invvar + z1_invvar).reciprocal()
    # average mean: u^T = (u1^T E1^-1 + u2^T E2^-1) E
    # disentanglement_lib: ave_mean = (z0_mean/z0_var + z1_mean/z1_var) * ave_var * 0.5
    ave_mean = (d0_posterior.mean*z0_invvar + d1_posterior.mean*z1_invvar) * ave_var * 0.5
    # done!
    return Normal(loc=ave_mean, scale=torch.sqrt(ave_var))


_COMPUTE_AVE_FNS = {
    'gvae': compute_average_gvae,
    'ml-vae': compute_average_ml_vae,
    'gvae_std': compute_average_gvae_std,  # this is un-official!
}


# ========================================================================= #
# Ada-GVAE                                                                  #
# ========================================================================= #


class AdaGVaeMinimal(BetaVae):
    """
    This is a direct implementation of the Ada-GVAE,
    which should be equivalent to the AdaVae with config values:

    >>> AdaVae.cfg(
    >>>    ada_average_mode='gvae',
    >>>    ada_thresh_mode='symmetric_kl',
    >>>    ada_thresh_ratio=0.5,
    >>> )
    """

    REQUIRED_OBS = 2

    def hook_intercept_ds(self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution]) -> Tuple[Sequence[Distribution], Sequence[Distribution], Dict[str, Any]]:
        """
        Adaptive VAE Method, putting the various components together
            1. compute differences between representations
            2. estimate a threshold for differences
            3. compute a shared mask from this threshold
            4. average together elements that are marked as shared

        (x) Visual inspection against reference implementation:
            https://github.com/google-research/disentanglement_lib (aggregate_argmax)
        """
        d0_posterior, d1_posterior = ds_posterior
        assert isinstance(d0_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d0_posterior)}'
        assert isinstance(d1_posterior, Normal), f'posterior distributions must be {Normal.__name__} distributions, got: {type(d1_posterior)}'

        # [1] symmetric KL Divergence FROM: https://openreview.net/pdf?id=8VXvj1QNRl1
        z_deltas = 0.5 * kl_divergence(d1_posterior, d0_posterior) + 0.5 * kl_divergence(d0_posterior, d1_posterior)

        # [2] estimate threshold from deltas
        z_deltas_min = z_deltas.min(axis=1, keepdim=True).values  # (B, 1)
        z_deltas_max = z_deltas.max(axis=1, keepdim=True).values  # (B, 1)
        z_thresh     = (0.5 * z_deltas_min + 0.5 * z_deltas_max)  # (B, 1)

        # [3] shared elements that need to be averaged, computed per pair in the batch
        share_mask = z_deltas < z_thresh  # broadcast (B, Z) and (B, 1) to get (B, Z)

        # [4.a] compute average representations
        # - this is the only difference between the Ada-ML-VAE
        ave_mean = (0.5 * d0_posterior.mean     + 0.5 * d1_posterior.mean)
        ave_std  = (0.5 * d0_posterior.variance + 0.5 * d1_posterior.variance) ** 0.5

        # [4.b] select shared or original values based on mask
        z0_mean = torch.where(share_mask, ave_mean, d0_posterior.loc)
        z1_mean = torch.where(share_mask, ave_mean, d1_posterior.loc)
        z0_std  = torch.where(share_mask, ave_std,  d0_posterior.scale)
        z1_std  = torch.where(share_mask, ave_std,  d1_posterior.scale)

        # construct distributions
        ave_d0_posterior = Normal(loc=z0_mean, scale=z0_std)
        ave_d1_posterior = Normal(loc=z1_mean, scale=z1_std)
        new_ds_posterior = (ave_d0_posterior, ave_d1_posterior)

        # [done] return new args & generate logs
        return new_ds_posterior, ds_prior, {
            'shared': share_mask.sum(dim=1).float().mean()
        }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

"""
The disent registry only contains implemented versions of
classes and functions, no interfaces are included.
  - for interfaces and creating your own data types, extend
    the base classes from the various locations.

*NB* All modules and classes are lazily imported!

You can register your own modules and classes using the provided decorator:
eg. `DATASET.register(...options...)(your_function_or_class)`
"""

# from disent.registry._registry import ProvidedValue
# from disent.registry._registry import StaticImport
# from disent.registry._registry import DictProviders
# from disent.registry._registry import RegexProvidersSearch

from disent.registry._registry import StaticValue
from disent.registry._registry import LazyValue
from disent.registry._registry import LazyImport
from disent.registry._registry import Registry
from disent.registry._registry import RegistryImports
from disent.registry._registry import RegexConstructor
from disent.registry._registry import RegexRegistry


# ========================================================================= #
# DATASETS - should be synchronized with: `disent/dataset/data/__init__.py` #
# ========================================================================= #


# TODO: this is not yet used in disent.data or disent.frameworks
DATASETS: RegistryImports['torch.utils.data.Dataset'] = RegistryImports('DATASETS')
# groundtruth -- impl
DATASETS['cars3d_x128']       = LazyImport('disent.dataset.data._groundtruth__cars3d.Cars3dData')
DATASETS['cars3d']            = LazyImport('disent.dataset.data._groundtruth__cars3d.Cars3d64Data')
DATASETS['dsprites']          = LazyImport('disent.dataset.data._groundtruth__dsprites.DSpritesData')
DATASETS['mpi3d_toy']         = LazyImport('disent.dataset.data._groundtruth__mpi3d.Mpi3dData', subset='toy')
DATASETS['mpi3d_realistic']   = LazyImport('disent.dataset.data._groundtruth__mpi3d.Mpi3dData', subset='realistic')
DATASETS['mpi3d_real']        = LazyImport('disent.dataset.data._groundtruth__mpi3d.Mpi3dData', subset='real')
DATASETS['smallnorb_x96']     = LazyImport('disent.dataset.data._groundtruth__norb.SmallNorbData')
DATASETS['smallnorb']         = LazyImport('disent.dataset.data._groundtruth__norb.SmallNorb64Data')
DATASETS['shapes3d']          = LazyImport('disent.dataset.data._groundtruth__shapes3d.Shapes3dData')
# groundtruth -- impl synthetic
DATASETS['xcolumns']          = LazyImport('disent.dataset.data._groundtruth__xcolumns.XColumnsData')
DATASETS['xyobject']          = LazyImport('disent.dataset.data._groundtruth__xyobject.XYObjectData')
DATASETS['xyobject_shaded']   = LazyImport('disent.dataset.data._groundtruth__xyobject.XYObjectShadedData')
DATASETS['xysquares']         = LazyImport('disent.dataset.data._groundtruth__xysquares.XYSquaresData')
DATASETS['xysquares_minimal'] = LazyImport('disent.dataset.data._groundtruth__xysquares.XYSquaresMinimalData')
DATASETS['xysinglesquare']    = LazyImport('disent.dataset.data._groundtruth__xysquares.XYSingleSquareData')


# ========================================================================= #
# SAMPLERS - should be synchronized with:                                   #
#            `disent/dataset/sampling/__init__.py`                          #
# ========================================================================= #


# TODO: this is not yet used in disent.data or disent.frameworks
# changes here should also update
SAMPLERS: RegistryImports['disent.dataset.sampling.BaseDisentSampler'] = RegistryImports('SAMPLERS')
# [ground truth samplers]
SAMPLERS['gt_dist']         = LazyImport('disent.dataset.sampling._groundtruth__dist.GroundTruthDistSampler')
SAMPLERS['gt_pair']         = LazyImport('disent.dataset.sampling._groundtruth__pair.GroundTruthPairSampler')
SAMPLERS['gt_pair_orig']    = LazyImport('disent.dataset.sampling._groundtruth__pair_orig.GroundTruthPairOrigSampler')
SAMPLERS['gt_single']       = LazyImport('disent.dataset.sampling._groundtruth__single.GroundTruthSingleSampler')
SAMPLERS['gt_triple']       = LazyImport('disent.dataset.sampling._groundtruth__triplet.GroundTruthTripleSampler')
# [any dataset samplers]
SAMPLERS['single']          = LazyImport('disent.dataset.sampling._single.SingleSampler')
SAMPLERS['random']          = LazyImport('disent.dataset.sampling._random__any.RandomSampler')
# [episode samplers]
SAMPLERS['random_episode']  = LazyImport('disent.dataset.sampling._random__episodes.RandomEpisodeSampler')


# ========================================================================= #
# FRAMEWORKS - should be synchronized with:                                 #
#             `disent/frameworks/ae/__init__.py`                            #
#             `disent/frameworks/ae/experimental/__init__.py`               #
#             `disent/frameworks/vae/__init__.py`                           #
#             `disent/frameworks/vae/experimental/__init__.py`              #
# ========================================================================= #


# TODO: this is not yet used in disent.frameworks
FRAMEWORKS: RegistryImports['disent.frameworks.DisentFramework'] = RegistryImports('FRAMEWORKS')
# [AE]
FRAMEWORKS['tae']           = LazyImport('disent.frameworks.ae._supervised__tae.TripletAe')
FRAMEWORKS['ae']            = LazyImport('disent.frameworks.ae._unsupervised__ae.Ae')
FRAMEWORKS['ada_ae']        = LazyImport('disent.frameworks.ae._weaklysupervised__adaae.AdaAe')             # ae version of ada_vae
FRAMEWORKS['adaneg_tae']    = LazyImport('disent.frameworks.ae._supervised__adaneg_tae.AdaNegTripletAe')    # ae version of adaneg_tvae
FRAMEWORKS['adaneg_tae_d']  = LazyImport('disent.frameworks.ae._unsupervised__dotae.DataOverlapTripletAe')  # ae version of adaneg_tvae_d
# [VAE]
FRAMEWORKS['tvae']          = LazyImport('disent.frameworks.vae._supervised__tvae.TripletVae')
FRAMEWORKS['betatc_vae']    = LazyImport('disent.frameworks.vae._unsupervised__betatcvae.BetaTcVae')
FRAMEWORKS['beta_vae']      = LazyImport('disent.frameworks.vae._unsupervised__betavae.BetaVae')
FRAMEWORKS['dfc_vae']       = LazyImport('disent.frameworks.vae._unsupervised__dfcvae.DfcVae')
FRAMEWORKS['dip_vae']       = LazyImport('disent.frameworks.vae._unsupervised__dipvae.DipVae')
FRAMEWORKS['info_vae']      = LazyImport('disent.frameworks.vae._unsupervised__infovae.InfoVae')
FRAMEWORKS['vae']           = LazyImport('disent.frameworks.vae._unsupervised__vae.Vae')
FRAMEWORKS['ada_vae']       = LazyImport('disent.frameworks.vae._weaklysupervised__adavae.AdaVae')
FRAMEWORKS['adaneg_tvae']   = LazyImport('disent.frameworks.vae._supervised__adaneg_tvae.AdaNegTripletVae')
FRAMEWORKS['adaneg_tvae_d'] = LazyImport('disent.frameworks.vae._unsupervised__dotvae.DataOverlapTripletVae')


# ========================================================================= #
# RECON_LOSSES - should be synchronized with:                               #
#                `disent/frameworks/helper/reconstructions.py`              #
# ========================================================================= #


RECON_LOSSES: RegexRegistry['disent.frameworks.helper.reconstructions.ReconLossHandler'] = RegexRegistry('RECON_LOSSES')  # TODO: we need a regex version of RegistryImports
# [STANDARD LOSSES]
RECON_LOSSES['mse']         = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerMse')                  # from the normal distribution - real values in the range [0, 1]
RECON_LOSSES['mae']         = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerMae')                  # mean absolute error
# [STANDARD DISTRIBUTIONS]
RECON_LOSSES['bce']         = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerBce')                  # from the bernoulli distribution - binary values in the set {0, 1}
RECON_LOSSES['bernoulli']   = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerBernoulli')            # reduces to bce - binary values in the set {0, 1}
RECON_LOSSES['c_bernoulli'] = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerContinuousBernoulli')  # bernoulli with a computed offset to handle values in the range [0, 1]
RECON_LOSSES['normal']      = LazyImport('disent.frameworks.helper.reconstructions.ReconLossHandlerNormal')               # handle all real values
# [REGEX LOSSES]
RECON_LOSSES.register_regex(pattern=r'^([a-z\d]+)_([a-z\d]+_[a-z\d]+)_l(\d+\.\d+)_k(\d+\.\d+)_norm_([a-z]+)$', example='mse_xy8_abs63_l1.0_k1.0_norm_none', factory_fn='disent.frameworks.helper.reconstructions._make_aug_recon_loss_l_w_n')
RECON_LOSSES.register_regex(pattern=r'^([a-z\d]+)_([a-z\d]+_[a-z\d]+)_norm_([a-z]+)$',                         example='mse_xy8_abs63_norm_none',           factory_fn='disent.frameworks.helper.reconstructions._make_aug_recon_loss_l1_w1_n')
RECON_LOSSES.register_regex(pattern=r'^([a-z\d]+)_([a-z\d]+_[a-z\d]+)$',                                       example='mse_xy8_abs63',                     factory_fn='disent.frameworks.helper.reconstructions._make_aug_recon_loss_l1_w1_nnone')


# ========================================================================= #
# LATENT_HANDLERS - should be synchronized with:                            #
#                  `disent/frameworks/helper/latent_distributions.py`       #
# ========================================================================= #


LATENT_HANDLERS: RegistryImports['disent.frameworks.helper.latent_distributions.LatentDistsHandler'] = RegistryImports('LATENT_HANDLERS')
LATENT_HANDLERS['normal']  = LazyImport('disent.frameworks.helper.latent_distributions.LatentDistsHandlerNormal')
LATENT_HANDLERS['laplace'] = LazyImport('disent.frameworks.helper.latent_distributions.LatentDistsHandlerLaplace')


# ========================================================================= #
# OPTIMIZER                                                                 #
# ========================================================================= #


# default learning rate for each optimizer
_LR = 1e-3

OPTIMIZERS: RegistryImports['torch.optim.Optimizer'] = RegistryImports('OPTIMIZERS')
# [torch]
OPTIMIZERS['adadelta']    = LazyImport(lr=_LR, import_path='torch.optim.adadelta.Adadelta')
OPTIMIZERS['adagrad']     = LazyImport(lr=_LR, import_path='torch.optim.adagrad.Adagrad')
OPTIMIZERS['adam']        = LazyImport(lr=_LR, import_path='torch.optim.adam.Adam')
OPTIMIZERS['adamax']      = LazyImport(lr=_LR, import_path='torch.optim.adamax.Adamax')
OPTIMIZERS['adam_w']      = LazyImport(lr=_LR, import_path='torch.optim.adamw.AdamW')
OPTIMIZERS['asgd']        = LazyImport(lr=_LR, import_path='torch.optim.asgd.ASGD')
OPTIMIZERS['lbfgs']       = LazyImport(lr=_LR, import_path='torch.optim.lbfgs.LBFGS')
OPTIMIZERS['rmsprop']     = LazyImport(lr=_LR, import_path='torch.optim.rmsprop.RMSprop')
OPTIMIZERS['rprop']       = LazyImport(lr=_LR, import_path='torch.optim.rprop.Rprop')
OPTIMIZERS['sgd']         = LazyImport(lr=_LR, import_path='torch.optim.sgd.SGD')
OPTIMIZERS['sparse_adam'] = LazyImport(lr=_LR, import_path='torch.optim.sparse_adam.SparseAdam')
# [torch_optimizer]
OPTIMIZERS['acc_sgd']     = LazyImport(lr=_LR, import_path='torch_optimizer.AccSGD')
OPTIMIZERS['ada_bound']   = LazyImport(lr=_LR, import_path='torch_optimizer.AdaBound')
OPTIMIZERS['ada_mod']     = LazyImport(lr=_LR, import_path='torch_optimizer.AdaMod')
OPTIMIZERS['adam_p']      = LazyImport(lr=_LR, import_path='torch_optimizer.AdamP')
OPTIMIZERS['agg_mo']      = LazyImport(lr=_LR, import_path='torch_optimizer.AggMo')
OPTIMIZERS['diff_grad']   = LazyImport(lr=_LR, import_path='torch_optimizer.DiffGrad')
OPTIMIZERS['lamb']        = LazyImport(lr=_LR, import_path='torch_optimizer.Lamb')
# 'torch_optimizer.Lookahead' is skipped because it is wrapped
OPTIMIZERS['novograd']    = LazyImport(lr=_LR, import_path='torch_optimizer.NovoGrad')
OPTIMIZERS['pid']         = LazyImport(lr=_LR, import_path='torch_optimizer.PID')
OPTIMIZERS['qh_adam']     = LazyImport(lr=_LR, import_path='torch_optimizer.QHAdam')
OPTIMIZERS['qhm']         = LazyImport(lr=_LR, import_path='torch_optimizer.QHM')
OPTIMIZERS['radam']       = LazyImport(lr=_LR, import_path='torch_optimizer.RAdam')
OPTIMIZERS['ranger']      = LazyImport(lr=_LR, import_path='torch_optimizer.Ranger')
OPTIMIZERS['ranger_qh']   = LazyImport(lr=_LR, import_path='torch_optimizer.RangerQH')
OPTIMIZERS['ranger_va']   = LazyImport(lr=_LR, import_path='torch_optimizer.RangerVA')
OPTIMIZERS['sgd_w']       = LazyImport(lr=_LR, import_path='torch_optimizer.SGDW')
OPTIMIZERS['sgd_p']       = LazyImport(lr=_LR, import_path='torch_optimizer.SGDP')
OPTIMIZERS['shampoo']     = LazyImport(lr=_LR, import_path='torch_optimizer.Shampoo')
OPTIMIZERS['yogi']        = LazyImport(lr=_LR, import_path='torch_optimizer.Yogi')


# ========================================================================= #
# METRIC - should be synchronized with: `disent/metrics/__init__.py`        #
# ========================================================================= #


# TODO: this is not yet used in disent.util.lightning.callbacks or disent.metrics
METRICS: RegistryImports['disent.metrics.utils._Metric'] = RegistryImports('METRICS')
METRICS['dci']                 = LazyImport('disent.metrics._dci.metric_dci')
METRICS['factor_vae']          = LazyImport('disent.metrics._factor_vae.metric_factor_vae')
METRICS['mig']                 = LazyImport('disent.metrics._mig.metric_mig')
METRICS['sap']                 = LazyImport('disent.metrics._sap.metric_sap')
METRICS['unsupervised']        = LazyImport('disent.metrics._unsupervised.metric_unsupervised')
# register metrics
METRICS['flatness']            = LazyImport('disent.metrics._flatness.metric_flatness')
METRICS['factored_components'] = LazyImport('disent.metrics._factored_components.metric_factored_components')
METRICS['distances']           = LazyImport('disent.metrics._factored_components.metric_distances')
METRICS['linearity']           = LazyImport('disent.metrics._factored_components.metric_linearity')


# ========================================================================= #
# SCHEDULE - should be synchronized with: `disent/schedule/__init__.py`     #
# ========================================================================= #


# TODO: this is not yet used in disent.framework or disent.schedule
SCHEDULES: RegistryImports['disent.schedule.Schedule'] = RegistryImports('SCHEDULES')
SCHEDULES['clip']        = LazyImport('disent.schedule._schedule.ClipSchedule')
SCHEDULES['cosine_wave'] = LazyImport('disent.schedule._schedule.CosineWaveSchedule')
SCHEDULES['cyclic']      = LazyImport('disent.schedule._schedule.CyclicSchedule')
SCHEDULES['linear']      = LazyImport('disent.schedule._schedule.LinearSchedule')
SCHEDULES['noop']        = LazyImport('disent.schedule._schedule.NoopSchedule')


# ========================================================================= #
# MODEL - should be synchronized with: `disent/model/ae/__init__.py`        #
# ========================================================================= #


# TODO: this is not yet used in disent.framework or disent.model
MODELS: RegistryImports['disent.model._base.DisentLatentsModule'] = RegistryImports('MODELS')
# [DECODER]
MODELS['encoder_conv64']     = LazyImport('disent.model.ae._vae_conv64.EncoderConv64')
MODELS['encoder_conv64norm'] = LazyImport('disent.model.ae._norm_conv64.EncoderConv64Norm')
MODELS['encoder_fc']         = LazyImport('disent.model.ae._vae_fc.EncoderFC')
MODELS['encoder_linear']     = LazyImport('disent.model.ae._linear.EncoderLinear')
# [ENCODER]
MODELS['decoder_conv64']     = LazyImport('disent.model.ae._vae_conv64.DecoderConv64')
MODELS['decoder_conv64norm'] = LazyImport('disent.model.ae._norm_conv64.DecoderConv64Norm')
MODELS['decoder_fc']         = LazyImport('disent.model.ae._vae_fc.DecoderFC')
MODELS['decoder_linear']     = LazyImport('disent.model.ae._linear.DecoderLinear')


# ========================================================================= #
# HELPER registries                                                         #
# ========================================================================= #


# TODO: add norm support with regex?
KERNELS: RegexRegistry['torch.Tensor'] = RegexRegistry('KERNELS')
KERNELS.register_regex(pattern=r'^box_r(\d+)$', example='box_r31', factory_fn='disent.dataset.transform._augment._make_box_kernel')
KERNELS.register_regex(pattern=r'^gau_r(\d+)$', example='gau_r31', factory_fn='disent.dataset.transform._augment._make_gaussian_kernel')


# ========================================================================= #
# Registry of all Registries                                                #
# ========================================================================= #


# registry of registries
REGISTRIES: Registry[Registry] = Registry('REGISTRIES')
REGISTRIES['DATASETS']        = StaticValue(DATASETS)
REGISTRIES['SAMPLERS']        = StaticValue(SAMPLERS)
REGISTRIES['FRAMEWORKS']      = StaticValue(FRAMEWORKS)
REGISTRIES['RECON_LOSSES']    = StaticValue(RECON_LOSSES)
REGISTRIES['LATENT_HANDLERS'] = StaticValue(LATENT_HANDLERS)
REGISTRIES['OPTIMIZERS']      = StaticValue(OPTIMIZERS)
REGISTRIES['METRICS']         = StaticValue(METRICS)
REGISTRIES['SCHEDULES']       = StaticValue(SCHEDULES)
REGISTRIES['MODELS']          = StaticValue(MODELS)
REGISTRIES['KERNELS']         = StaticValue(KERNELS)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import sys
from multiprocessing import current_process
from typing import Optional

import signal
import logging

from pytorch_lightning import Trainer
from pytorch_lightning.loggers import LoggerCollection

from disent.util.lightning.logger_util import wb_yield_loggers

log = logging.getLogger(__name__)


# ========================================================================= #
# HYDRA CONFIG HELPERS                                                      #
# ========================================================================= #

_PL_SIGNALS_OLD_HANDLERS = {}
_PL_SIGNALS = (      # we can't capture SIGKILL
    signal.SIGINT,   # interrupted from the dialogue station
    signal.SIGTERM,  # terminate the process in a soft way
    signal.SIGABRT,  # abnormal termination
    signal.SIGSEGV,  # segmentation fault
)

_PL_LOGGER: Optional[LoggerCollection] = None
_PL_TRAINER: Optional[Trainer] = None


def safe_unset_debug_trainer():
    global _PL_TRAINER
    if _PL_TRAINER is not None:
        _PL_TRAINER = None


def set_debug_trainer(trainer: Optional[Trainer]):
    global _PL_TRAINER
    assert _PL_TRAINER is None, 'debug trainer has already been set'
    _PL_TRAINER = trainer
    return trainer


def _signal_handler_log_and_exit(signal_number, frame):
    # call in all the child processes for the best chance of clearing this...
    # remove callbacks from trainer so we aren't stuck running forever!
    # TODO: this is a hack... there must be a better way to do this... could it be a pl bug?
    #       this logic is duplicated in the framework training_step
    if _PL_TRAINER and _PL_TRAINER.callbacks:
        _PL_TRAINER.callbacks.clear()

    # make sure that we only exit in the parent process
    if current_process().name != 'MainProcess':
        log.debug('Skipping signal handling for child process!')
        return
    # get the signal name
    numbers_to_names = dict((k, v) for v, k in reversed(sorted(signal.__dict__.items())) if v.startswith('SIG') and not v.startswith('SIG_'))
    signal_name = numbers_to_names.get(signal_number, signal_number)
    # log everything!
    log_error_and_exit(
        err_type=f'received exit signal',
        err_msg=f'{signal_name}',
        exit_code=signal_number,
        exc_info=False,
    )


def safe_unset_debug_logger():
    global _PL_LOGGER
    # unset logger
    if _PL_LOGGER is not None:
        _PL_LOGGER = None
        # return control to original handlers
        for signal_type in _PL_SIGNALS:
            if signal_type in _PL_SIGNALS_OLD_HANDLERS:
                handler = _PL_SIGNALS_OLD_HANDLERS.pop(signal_type)
                signal.signal(signal_type, handler)


def set_debug_logger(logger: Optional[LoggerCollection]):
    global _PL_LOGGER
    assert _PL_LOGGER is None, 'debug logger has already been set'
    _PL_LOGGER = logger
    # set initial messages
    if _PL_LOGGER is not None:
        _PL_LOGGER.log_metrics({
            'error_type': 'N/A',
            'error_msg': 'N/A',
            'error_occurred': False,
        })
    # register signal listeners
    for signal_type in _PL_SIGNALS:
        # save the old handler
        _PL_SIGNALS_OLD_HANDLERS[signal_type] = signal.getsignal(signal_type)
        # update the handler
        signal.signal(signal_type, _signal_handler_log_and_exit)
    # return the logger
    return logger


def log_error_and_exit(err_type: str, err_msg: str, exit_code: int = 1, exc_info=True):
    # truncate error
    err_msg = err_msg[:244] + ' <TRUNCATED>' if len(err_msg) > 244 else err_msg
    # log something at least
    log.error(f'exiting: {err_type} | {err_msg}', exc_info=exc_info)
    # try log to pytorch lightning & wandb
    if _PL_LOGGER is not None:
        _PL_LOGGER.log_metrics({
            'error_type': err_type,
            'error_msg': err_msg,
            'error_occurred': True,
        })
        for wb_logger in wb_yield_loggers(_PL_LOGGER):
            # so I dont have to scroll up... I'm lazy...
            run_url = wb_logger.experiment._get_run_url()
            project_url = wb_logger.experiment._get_project_url()
            log.error(f'wandb: run url: {run_url if run_url else "N/A"}')
            log.error(f'wandb: project url: {project_url if run_url else "N/A"}')
            # make sure we log everything online!
            wb_logger.experiment.finish(exit_code=exit_code)
    # EXIT!
    sys.exit(exit_code)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
